import abc
import math
import os
from collections import defaultdict
from pathlib import Path
from typing import Tuple, Dict, Any, List

import aiofiles
from gcp_pilot.datastore import Document, DoesNotExist
from gcp_pilot.exceptions import ValidationError
from sanic.request import Request, RequestParameters, File
from sanic.response import json, HTTPResponse
from sanic.views import HTTPMethodView

from sanic_rest import exceptions, validator

PayloadType = Dict[str, Any]
ResponseType = Tuple[PayloadType, int]

STAGE_DIR = os.environ.get('SANIC_FILE_DIR', default=Path(__file__).parent)


class ViewBase(HTTPMethodView):
    model: Document

    @classmethod
    async def write_file(cls, file: File, filepath: Path):
        filepath.parent.mkdir(parents=True, exist_ok=True)
        async with aiofiles.open(filepath, 'wb') as f:
            await f.write(file.body)
        await f.close()

    def validate(self, data, model_klass=None, partial=False):
        model_klass = model_klass or self.model
        info = validator.ModelInfo.build(model_klass=model_klass)
        errors = info.validate(item=data, partial=partial)
        if errors:
            raise exceptions.ValidationError(message=errors)

    def _validate_item(self, item, options, partial=False) -> Dict[str, Any]:
        errors = {}

        for field_name, field_info in options['fields']:
            if not partial and field_info['required'] and field_name not in item:
                errors[field_name] = f"Mandatory field"
                continue

            provided_items = item[field_name]

            if field_info['repeated']:
                for provided_item in provided_items:
                    errors[field_name] = self._validate_item(item=provided_item, options=field_info, partial=False)

        return errors


class ListView(ViewBase, abc.ABC):
    search_field: str = None

    async def get(self, request: Request) -> HTTPResponse:
        query_args, page, page_size = self._parse_query_args(request=request)
        try:
            items = await self.perform_get(query_filters=query_args)
        except ValidationError as e:
            raise exceptions.ValidationError(message=str(e))

        items_in_page = self._paginate(items=items, page=page, page_size=page_size)

        response = {
            'results': [
                obj.serialize()
                for obj in
                items_in_page
            ],
            'count': len(items),
            'num_pages': int(math.ceil(len(items) / page_size))
        }
        return json(response, 200 if any(items_in_page) else 404)

    def _parse_query_args(self, request: Request) -> Tuple[Dict[str, Any], int, int]:
        query_args = dict()
        for key, value in request.query_args:
            if key == 'q':
                key = f'{self.search_field}__startswith'

            if key not in query_args:
                query_args[key] = value
            elif isinstance(query_args[key], list):
                query_args[key].append(value)
            else:
                query_args[key] = [query_args[key], value]

        # Fetch & validate pagination params
        try:
            page = int(query_args.pop('page', 1))
            page_size = int(query_args.pop('page_size', 10))
        except ValueError as e:
            raise exceptions.ValidationError(f"page and page_size must be integers") from e
        if page < 1:
            raise exceptions.ValidationError(f"page starts at 1")
        if page_size < 1:
            raise exceptions.ValidationError(f"page_size must be at least at 1")

        return query_args, page, page_size

    def _paginate(self, items: List[Document], page: int, page_size: int) -> List[Document]:
        # TODO Add proper pagination with cursors
        start_idx = (page - 1) * page_size
        start_idx = min(start_idx, len(items))

        end_idx = start_idx + page_size
        end_idx = min(end_idx, len(items))

        items_in_page = items[start_idx:end_idx]
        return items_in_page

    async def perform_get(self, query_filters) -> List[Document]:
        return list(self.model.documents.filter(**query_filters))

    async def post(self, request: Request) -> HTTPResponse:
        data = request.json
        self.validate(data=data, partial=False)

        try:
            obj = await self.perform_create(data=data)
        except ValidationError as e:
            raise exceptions.ValidationError(message=str(e))

        data = obj.serialize()
        return json(data, 201)

    async def perform_create(self, data: PayloadType) -> Document:
        obj = self.model.deserialize(**data)
        return obj.save()

    async def options(self, request: Request) -> HTTPResponse:
        data = await self.perform_options()
        return json(data, 200)

    async def perform_options(self) -> PayloadType:
        return validator.ModelInfo.build(model_klass=self.model).as_dict


class DetailView(ViewBase, abc.ABC):
    def _parse_pk(self, pk: str):
        pk_field_name = self.model.Meta.pk_field
        pk_field_type = self.model.Meta.fields[self.model.Meta.pk_field]
        try:
            return pk_field_type(pk)
        except ValueError:
            raise exceptions.ValidationError(f"{pk_field_name} field must be {pk_field_type.__name__}")

    async def get(self, request: Request, pk: str) -> HTTPResponse:
        try:
            obj = await self.perform_get(pk=self._parse_pk(pk=pk), query_filters={})
        except DoesNotExist as e:
            raise exceptions.NotFoundError() from e
        except ValidationError as e:
            raise exceptions.ValidationError(message=str(e))

        data = obj.serialize()
        return json(data, 200)

    async def perform_get(self, pk, query_filters) -> Document:
        return self.model.documents.get(id=pk, **query_filters)

    async def put(self, request: Request, pk: str) -> HTTPResponse:
        payload = request.json
        self.validate(data=payload, partial=True)

        try:
            obj = await self.perform_create(data=payload)
        except ValidationError as e:
            raise exceptions.ValidationError(message=str(e))

        data = obj.serialize()
        return json(data, 200)

    async def perform_create(self, data: PayloadType) -> Document:
        obj = self.model.deserialize(**data)
        return obj.save()

    async def patch(self, request: Request, pk: str) -> HTTPResponse:
        if request.files:
            payload = {}
        else:
            payload = request.json
            self.validate(data=payload, partial=True)

        try:
            obj = await self.perform_update(pk=self._parse_pk(pk=pk), data=payload, files=request.files)
        except ValidationError as e:
            raise exceptions.ValidationError(message=str(e))

        data = obj.serialize()
        return json(data, 200)

    async def perform_update(self, pk: str, data: PayloadType, files: RequestParameters) -> Document:
        obj = self.model.documents.update(pk=pk, **data)

        file_updates: Dict[str, Any] = defaultdict(list)
        for key, file in files.items():
            filepath = await self.store_file(
                obj=obj,
                field_name=key,
                file=file[0],  # TODO: check if it's possible to have more files
            )
            if '__' not in key:
                file_updates[key] = filepath
            else:
                key = key.split('__')[0]
                file_updates[key].append(filepath)
        obj = self.model.documents.update(pk=pk, **file_updates)

        return obj

    async def store_file(self, obj: Document, field_name: str, file: File) -> str:
        local_filepath = STAGE_DIR / 'media' / obj.pk / field_name
        await self.write_file(file=file, filepath=local_filepath)
        return str(local_filepath)

    async def delete(self, request: Request, pk: str) -> HTTPResponse:
        await self.perform_delete(pk=pk)
        return json({}, 204)

    async def perform_delete(self, pk: str) -> None:
        self.model.documents.delete(pk=pk)


class ActionView(ViewBase):
    def get_model(self, pk: str) -> Document:
        return self.model.documents.get(id=pk)

    async def get(self, request: Request, pk: str) -> HTTPResponse:
        try:
            obj = self.get_model(pk=pk)
        except DoesNotExist as e:
            raise exceptions.NotFoundError() from e

        try:
            data, status = await self.perform_get(request=request, obj=obj)
        except ValidationError as e:
            raise exceptions.ValidationError(message=str(e))

        return json(data, status)

    @abc.abstractmethod
    async def perform_get(self, request: Request, obj: Document) -> ResponseType:
        raise NotImplementedError()

    async def post(self, request: Request, pk: str) -> HTTPResponse:
        try:
            obj = self.get_model(pk=pk)
        except DoesNotExist as e:
            raise exceptions.NotFoundError() from e

        try:
            data, status = await self.perform_post(request=request, obj=obj)
        except ValidationError as e:
            raise exceptions.ValidationError(message=str(e))

        return json(data, status)

    @abc.abstractmethod
    async def perform_post(self, request: Request, obj: Document) -> ResponseType:
        raise NotImplementedError()

    async def delete(self, request: Request, pk: str) -> HTTPResponse:
        try:
            obj = self.get_model(pk=pk)
        except DoesNotExist as e:
            raise exceptions.NotFoundError() from e

        data, status = await self.perform_delete(request=request, obj=obj)
        return json(data, status)

    @abc.abstractmethod
    async def perform_delete(self, request: Request, obj: Document) -> ResponseType:
        raise NotImplementedError()



from abc import ABC

from sanic_jwt import protected

from sanic_rest.views import ListView, DetailView, ActionView


class ProtectedMixin:
    decorators = [protected()]


class ProtectedListView(ProtectedMixin, ListView, ABC):
    pass


class ProtectedDetailView(ProtectedMixin, DetailView, ABC):
    pass


class ProtectedActionView(ProtectedMixin, ActionView, ABC):
    pass



from enum import Enum
import inspect
from dataclasses import _MISSING_TYPE, dataclass
from typing import Any, List, Dict, _GenericAlias, Union

from gcp_pilot.datastore import EmbeddedDocument

AnyInfoType = Union['FieldInfo', 'ModelInfo']


@dataclass
class FieldInfo:
    type: str
    required: bool

    @classmethod
    def build(cls, field_klass, required: bool) -> 'FieldInfo':
        if inspect.isclass(field_klass):
            field_type = field_klass.__name__
            if issubclass(field_klass, EmbeddedDocument):
                return ModelInfo.build(model_klass=field_klass, required=required)
            if issubclass(field_klass, Enum):
                return OptionsFieldInfo(type=field_type, required=required, choices=[e.value for e in field_klass])
            return FieldInfo(type=field_type, required=required)
        elif isinstance(field_klass, _GenericAlias):
            field_type = field_klass._name
            if field_type == 'List':
                inner_klass = field_klass.__args__[0]
                return ListFieldInfo(
                    type=field_type,
                    required=required,
                    struct=cls.build(field_klass=inner_klass, required=True),
                )
            elif field_type == 'Tuple':
                return TupleFieldInfo(
                    type=field_type,
                    required=required,
                    structs=[
                        cls.build(field_klass=inner_klass, required=True)
                        for inner_klass in field_klass.__args__
                    ],
                )
            elif field_type == 'Dict':
                key_klass, value_klass = field_klass.__args__
                return DictFieldInfo(
                    type=field_type,
                    required=required,
                    key_struct=cls.build(field_klass=key_klass, required=True),
                    value_struct=cls.build(field_klass=value_klass, required=True),
                )
            else:
                raise RuntimeError(f"Unable to handle field class {field_klass}")
        elif field_klass == Any:
            return FieldInfo(type='any', required=required)
        else:
            raise RuntimeError(f"Unable to handle field class {field_klass}")

    def validate(self, item: Dict[Any, Any]) -> Union[str, None]:
        if self.required and item is None:
            return "Must not be empty"
        if isinstance(item, dict) and self.type != 'dict':
            return f"Must be a {self.type}"
        return None

    @property
    def as_dict(self) -> Dict[str, Any]:
        return {
            'type': self.type,
            'required': self.required,
        }


@dataclass
class OptionsFieldInfo(FieldInfo):
    choices: List[Any]

    @property
    def as_dict(self) -> Dict[str, Any]:
        data = super().as_dict
        data['choices'] = self.choices
        return data

    def validate(self, item: Dict[Any, Any]) -> Union[str, None]:
        error = super().validate(item=item)
        if error:
            return error

        if item not in self.choices:
            return f"It must be one of {self.choices}. Not {item}"

        return None


@dataclass
class ModelInfo(FieldInfo):
    fields: Dict[str, FieldInfo]

    @classmethod
    def build(cls, model_klass: EmbeddedDocument, required=True) -> 'ModelInfo':
        hints = model_klass.__dataclass_fields__

        fields = {}
        for field_name, _ in model_klass.Meta.fields.items():
            hint = hints[field_name]
            field_required = isinstance(hint.default, _MISSING_TYPE) and isinstance(hint.default_factory, _MISSING_TYPE)
            field_klass = hint.type

            info = FieldInfo.build(field_klass=field_klass, required=field_required)
            fields[field_name] = info
        return cls(type=model_klass.__name__, fields=fields, required=required)

    def validate(self, item: Dict[Any, Any], partial: bool = False) -> Union[str, Dict[str, Any]]:
        errors = {}

        if not isinstance(item, dict):
            return f"Must be key-value ({self.type})"

        for field_name, field_info in self.fields.items():
            if field_name not in item:
                if not partial and field_info.required:
                    errors[field_name] = f"This field is required"
                continue

            field_errors = field_info.validate(item=item[field_name])
            if field_errors:
                errors[field_name] = field_errors

        return errors

    @property
    def as_dict(self):
        data = super().as_dict
        data['fields'] = {key: value.as_dict for key, value in self.fields.items()}
        return data


@dataclass
class ListFieldInfo(FieldInfo):
    struct: AnyInfoType

    def validate(self, item):
        if not isinstance(item, list):
            return 'Must be a list'

        errors = {}
        for idx, inner_item in enumerate(item):
            inner_error = self.struct.validate(item=inner_item)
            if inner_error:
                errors[f'#{idx}'] = inner_error
        return errors

    @property
    def as_dict(self) -> Dict[str, Any]:
        data = super().as_dict
        data['struct'] = self.struct.as_dict
        return data


@dataclass
class TupleFieldInfo(FieldInfo):
    structs: List[AnyInfoType]

    def validate(self, item):
        if not isinstance(item, (list, tuple)):
            return 'Must be a list'
        elif len(item) != len(self.structs):
            return f'Must be a list of {len(self.structs)} items'

        errors = {}
        for idx, (inner_item, struct) in enumerate(zip(item, self.structs)):
            inner_error = struct.validate(item=inner_item)
            if inner_error:
                errors[f'#{idx}'] = inner_error
        return errors

    @property
    def as_dict(self) -> Dict[str, Any]:
        data = super().as_dict
        data['structs'] = [struct.as_dict for struct in self.structs]
        return data


@dataclass
class DictFieldInfo(FieldInfo):
    key_struct: AnyInfoType
    value_struct: AnyInfoType

    def validate(self, item):
        if not isinstance(item, dict):
            return 'Must be key-value'

        errors = {}
        for idx, (item_key, item_value) in enumerate(item.items()):
            item_error = self.key_struct.validate(item_key)
            if item_error:
                errors[f'#{idx}_KEY'] = item_error

            value_error = self.value_struct.validate(item_value)
            if value_error:
                errors[f'#{idx}_VALUE'] = value_error
        return errors

    @property
    def as_dict(self) -> Dict[str, Any]:
        data = super().as_dict
        data['key_struct'] = self.key_struct.as_dict
        data['value_struct'] = self.value_struct.as_dict
        return data



from dataclasses import dataclass

from sanic.handlers import ErrorHandler
from sanic.response import json


@dataclass
class HttpException(Exception):
    message: str
    status_code: int

    @property
    def response(self):
        return {'error': self.message}


@dataclass
class ValidationError(HttpException):
    status_code: int = 400


@dataclass
class NotFoundError(HttpException):
    message: str = "Not found"
    status_code: int = 404


@dataclass
class NotAllowedError(HttpException):
    message: str = "Not allowed"
    status_code: int = 405


@dataclass
class ForbiddenError(HttpException):
    message: str = "Forbidden"
    status_code: int = 409


def _http_error_handler(request, exc):
    return json(exc.response, exc.status_code)


rest_error_handler = ErrorHandler()
rest_error_handler.add(exception=HttpException, handler=_http_error_handler)
